from ftw.raven.browser.demo import RavenConnectionTest
from ftw.raven.client import get_raven_client
from ftw.raven.tests import FunctionalTestCase
from plone.app.testing import SITE_OWNER_NAME
import os


class TestIntegration(FunctionalTestCase):

    def test_errors_in_requests_are_captured(self):
        os.environ['RAVEN_DSN'] = 'https://x:y@sentry.local/1'
        self.request_to_error_view(view='raven-test', login_as=SITE_OWNER_NAME)
        calls = get_raven_client().captureException_calls
        self.assertEquals(1, len(calls), 'Expected one raven client call')
        self.assertEquals(RavenConnectionTest,
                          calls[0]['exc_info'][0],
                          'Expected a RavenConnectionTest to be reported.')



from ftw.raven.config import get_raven_config
from ftw.raven.tests import FunctionalTestCase
import os
import tempfile


class TestRavenConfig(FunctionalTestCase):

    def test_not_enabled_when_not_configured(self):
        self.assertFalse(get_raven_config().enabled)

    def test_enabled_when_dsn_configured(self):
        os.environ['RAVEN_DSN'] = 'https://123:456@sentry.local/2'
        self.assertTrue(get_raven_config().enabled)

    def test_dsn_is_accessible_as_attribute(self):
        os.environ['RAVEN_DSN'] = 'https://123:456@sentry.local/3'
        self.assertEquals('https://123:456@sentry.local/3',
                          get_raven_config().dsn)

    def test_project_dist(self):
        os.environ['RAVEN_PROJECT_DIST'] = 'my.project'
        self.assertEquals('my.project', get_raven_config().project_dist)

    def test_buildout_root(self):
        parent_dir = os.path.dirname(__file__)
        os.environ['RAVEN_BUILDOUT_ROOT'] = parent_dir
        self.assertEquals(parent_dir, get_raven_config().buildout_root)

    def test_ignored_exception_classnames(self):
        self.assertEquals(
            set(['Redirect', 'NotFound', 'Unauthorized', 'Intercepted']),
            set(get_raven_config().ignored_exception_classnames))

        os.environ['RAVEN_ENABLE_EXCEPTIONS'] = 'NotFound'
        self.assertEquals(
            set(['Redirect', 'Unauthorized', 'Intercepted']),
            set(get_raven_config().ignored_exception_classnames))

        os.environ['RAVEN_ENABLE_EXCEPTIONS'] = 'Unauthorized, Redirect'
        self.assertEquals(
            set(['NotFound', 'Intercepted']),
            set(get_raven_config().ignored_exception_classnames))

        os.environ['RAVEN_ENABLE_EXCEPTIONS'] = 'Unauthorized, Redirect, NotFound'
        self.assertEquals(
            set(['Intercepted']),
            set(get_raven_config().ignored_exception_classnames))

        os.environ['RAVEN_ENABLE_EXCEPTIONS'] = 'Unauthorized, Redirect, NotFound, Intercepted'
        self.assertEquals(
            set(),
            set(get_raven_config().ignored_exception_classnames))

    def test_disabling_custom_exceptions(self):
        self.assertNotIn('RandomError',
                         get_raven_config().ignored_exception_classnames)

        os.environ['RAVEN_DISABLE_EXCEPTIONS'] = 'RandomError'
        self.assertIn('RandomError',
                      get_raven_config().ignored_exception_classnames)

        os.environ['RAVEN_DISABLE_EXCEPTIONS'] = 'Foo, RandomError, Bar'
        self.assertIn('RandomError',
                      get_raven_config().ignored_exception_classnames)

    def test_no_tags_by_default(self):
        self.assertEquals({}, get_raven_config().tags)

    def test_custom_tags_with_env_variable(self):
        os.environ['RAVEN_TAGS'] = ('{"deployment_type": "production",'
                                    ' "maintainer": "peter"}')
        self.assertEquals({"deployment_type": "production",
                           "maintainer": "peter"},
                          get_raven_config().tags)

    def test_custom_tags_may_be_invalid(self):
        os.environ['RAVEN_TAGS'] = '{"something completely wrong":}'
        self.assertEquals({'RAVEN_TAGS': 'invalid configuration'},
                          get_raven_config().tags)

    def test_custom_tags_by_file(self):
        with tempfile.NamedTemporaryFile() as tags_file:
            os.environ['RAVEN_TAGS_FILE'] = tags_file.name
            tags_file.write('{"status": "testing"}')
            tags_file.seek(0)

            self.assertEquals({'status': 'testing'},
                              get_raven_config().tags)

    def test_file_content_may_be_invalid(self):
        with tempfile.NamedTemporaryFile() as tags_file:
            os.environ['RAVEN_TAGS_FILE'] = tags_file.name
            self.assertEquals({'RAVEN_TAGS_FILE': 'invalid content'},
                              get_raven_config().tags)

            tags_file.write('not json at all')
            tags_file.seek(0)
            self.assertEquals({'RAVEN_TAGS_FILE': 'invalid content'},
                              get_raven_config().tags)

    def test_file_may_not_exist(self):
        os.environ['RAVEN_TAGS_FILE'] = '/foo/br'
        self.assertEquals({'RAVEN_TAGS_FILE': 'missing file'},
                          get_raven_config().tags)

    def test_combining_tags_from_file_and_env_variable(self):
        os.environ['RAVEN_TAGS'] = '{"maintainer": "hans", "priority": "env"}'
        with tempfile.NamedTemporaryFile() as tags_file:
            os.environ['RAVEN_TAGS_FILE'] = tags_file.name
            tags_file.write('{"purpose": "demo", "priority": "file"}')
            tags_file.seek(0)

            self.assertEquals({'purpose': 'demo',
                               'maintainer': 'hans',
                               'priority': 'env'},
                              get_raven_config().tags)



from ftw.raven.config import get_raven_config
import raven


raven_client_class = raven.Client
_client_cache = None


def get_raven_client():
    config = get_raven_config()
    if not config.enabled:
        return None

    if globals().get('_client_cache', None) is not None:
        return globals()['_client_cache']

    globals()['_client_cache'] = raven_client_class(
        dsn=config.dsn,
        install_sys_hook=False)
    return globals()['_client_cache']


def purge_raven_client():
    globals()['_client_cache'] = None



from ftw.raven.client import get_raven_client
from ftw.raven.client import purge_raven_client
from ftw.raven.tests import FunctionalTestCase
import os


class TestClient(FunctionalTestCase):

    def test_no_client_when_not_configured(self):
        self.assertIsNone(get_raven_client())

    def test_client_is_utility(self):
        os.environ['RAVEN_DSN'] = dsn = 'https://foo:bar@sentry.local/3'
        self.assertEquals(dsn, get_raven_client().dsn)

    def test_client_is_cached_and_can_be_purged(self):
        os.environ['RAVEN_DSN'] = 'https://x:y@sentry.local/1'
        first_client = get_raven_client()
        self.assertTrue(first_client)
        self.assertEquals(id(first_client), id(get_raven_client()))
        purge_raven_client()
        self.assertNotEquals(id(first_client), id(get_raven_client()))



from ftw.builder import Builder
from ftw.builder import create
from ftw.raven.client import get_raven_client
from ftw.raven.tests import FunctionalTestCase
from ftw.raven.tests.client_mock import CrashingClientMock
from pkg_resources import get_distribution
import os


class TestIntegration(FunctionalTestCase):

    def test_errors_in_requests_are_captured(self):
        os.environ['RAVEN_DSN'] = 'https://x:y@sentry.local/1'
        self.request_to_error_view()
        calls = get_raven_client().captureException_calls
        self.assertEquals(1, len(calls), 'Expected one raven client call')
        self.assertEquals(KeyError, calls[0]['exc_info'][0],
                          'Expected a KeyError to be reported.')

    def test_request_infos(self):
        call = self.make_error_and_get_capture_call(data={'foo': '3'})
        self.maxDiff = None

        expected = call['data']['request']

        # "X-Theme-Enabled" is available on Plone 5 but not on Plone 4.
        # It's irrelevant for our test so we can remove it.
        expected['headers'].pop('X-Theme-Enabled', None)

        self.assertEquals(
            {'cookies': {},
             'url': 'http://nohost/plone/make_key_error',
             'headers': {'Http-Referer': '',
                         'Connection': 'close',
                         'Referer': '',
                         'Host': 'nohost',
                         'User-Agent': 'Python-urllib/2.7'},
             'env': {'CONTENT_LENGTH': '5',
                     'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                     'QUERY_STRING': '',
                     'REQUEST_METHOD': 'POST',
                     'PATH_INFO': '/plone/make_key_error',
                     'SERVER_PROTOCOL': 'HTTP/1.1'},
             'query_string': '',
             'data': {'foo': '3'},
             'method': 'POST'},
            expected)

    def test_user_infos_as_anonymous(self):
        call = self.make_error_and_get_capture_call()
        self.maxDiff = None
        self.assertEquals(
            {'authentication': 'anonymous',
             'ip_address': '',
             'roles': ('Anonymous',),
             'roles_in_context': ['Anonymous']},
            call['data']['user'])

    def test_user_infos_as_logged_in_user(self):
        user = create(Builder('user'))
        call = self.make_error_and_get_capture_call(login_as=user)
        self.maxDiff = None
        self.assertEquals(
            {'authentication': 'authenticated',
             'email': 'john@doe.com',
             'fullname': 'Doe John',
             'id': 'john.doe',
             'username': 'john.doe',
             'ip_address': '',
             'roles': ['Member', 'Authenticated'],
             'roles_in_context': ['Member', 'Authenticated']},
            call['data']['user'])

    def test_extra_infos(self):
        call = self.make_error_and_get_capture_call()
        self.assertIn('context', call['data']['extra'])
        self.assertIn('request.other', call['data']['extra'])

    def test_modules_infos(self):
        call = self.make_error_and_get_capture_call()
        self.assertIn('ftw.raven', call['data']['modules'])
        self.assertIn('python', call['data']['modules'])

    def test_release_not_sent_when_nothing_configured(self):
        call = self.make_error_and_get_capture_call()
        self.assertNotIn('release', call['data'])

    def test_release_with_project_dist(self):
        os.environ['RAVEN_PROJECT_DIST'] = 'zope.interface'
        call = self.make_error_and_get_capture_call()
        version = get_distribution('zope.interface').version
        self.assertDictContainsSubset({'release': version}, call['data'])

    def test_exceptions_catched_when_client_crashes(self):
        CrashingClientMock.install()
        os.environ['RAVEN_DSN'] = 'https://x:y@sentry.local/1'
        self.assertEquals(0, get_raven_client().crashes)
        self.request_to_error_view()
        self.assertEquals(
            2, get_raven_client().crashes,
            'We excpect exactly two attempts to report an exception:'
            ' 1. Report the actual exception,'
            ' 2. Report the meta exception that the first one failed.')

    def test_404_reporting_can_be_enabled(self):
        os.environ['RAVEN_DSN'] = 'https://x:y@sentry.local/1'
        self.request_to_error_view(view='make_404_err')
        self.assertEquals(0, len(get_raven_client().captureException_calls))

        os.environ['RAVEN_ENABLE_EXCEPTIONS'] = 'NotFound'
        self.request_to_error_view(view='make_404_err')
        self.assertEquals(1, len(get_raven_client().captureException_calls))

    def test_error_log_id_is_reported_as_tag(self):
        call = self.make_error_and_get_capture_call()
        self.assertIn('error_log_id', call['data']['tags'])

    def test_additional_tags_are_reported(self):
        os.environ['RAVEN_TAGS'] = '{"maintainer": "hugo"}'
        call = self.make_error_and_get_capture_call()
        self.assertDictContainsSubset({'maintainer': 'hugo'},
                                      call['data']['tags'])

    def make_error_and_get_capture_call(self, **kwargs):
        os.environ['RAVEN_DSN'] = 'https://x:y@sentry.local/1'
        self.request_to_error_view(**kwargs)
        calls = get_raven_client().captureException_calls
        self.assertEquals(1, len(calls), 'Expected one raven client call')
        return calls[0]



from ftw.raven import reporter
from ftw.raven.client import get_raven_client
from ftw.raven.tests import FunctionalTestCase
from pkg_resources import get_distribution
from zExceptions import NotFound
from zExceptions import Redirect
from zExceptions import Unauthorized
import os


class TestReporter(FunctionalTestCase):

    def test_silent_handling_of_request_preparation_exceptions(self):
        """When an exception happens while preparing the request,
        the exception should be catched and reported as raven_meta_error.
        """
        os.environ['RAVEN_DSN'] = 'https://x:y@sentry.local/1'
        self.assertEquals(0, len(get_raven_client().captureException_calls))
        reporter.maybe_report_exception(None, None, KeyError, None, None)
        self.assertEquals(1, len(get_raven_client().captureException_calls),
                          'Expected exactly one error to be reported.')
        call, = get_raven_client().captureException_calls
        self.assertIn('raven_meta_error', call['data']['extra'])

    def test_release_is_None_when_nothing_configured(self):
        self.assertIsNone(reporter.get_release())

    def test_release_can_be_project_dist_version(self):
        os.environ['RAVEN_PROJECT_DIST'] = 'zope.interface'
        version = get_distribution('zope.interface').version
        self.assertEquals(version, reporter.get_release())

    def test_no_exception_when_dist_not_existing(self):
        os.environ['RAVEN_PROJECT_DIST'] = 'foo.bar'
        self.assertIsNone(reporter.get_release())

    def test_release_can_be_buildout_git_hash(self):
        os.environ['RAVEN_BUILDOUT_ROOT'] = os.path.abspath(os.path.join(
            __file__, '..', '..', '..', '..'))
        release = reporter.get_release()
        self.assertTrue(release)
        self.assertEquals(40, len(release),
                          'Expected git SHA to be 40 bytes big.')

    def test_is_exception_type_ignored_can_handle_None(self):
        self.assertTrue(reporter.is_exception_type_ignored(None))

    def test_default_exceptions_are_not_ignored(self):
        self.assertFalse(reporter.is_exception_type_ignored(KeyError))
        self.assertFalse(reporter.is_exception_type_ignored(Exception))

    def test_default_ignored_exceptions(self):
        self.assertTrue(reporter.is_exception_type_ignored(NotFound))
        self.assertTrue(reporter.is_exception_type_ignored(Unauthorized))
        self.assertTrue(reporter.is_exception_type_ignored(Redirect))

    def test_enabling_default_ignored_exceptions(self):
        os.environ['RAVEN_ENABLE_EXCEPTIONS'] = 'NotFound'
        self.assertFalse(reporter.is_exception_type_ignored(NotFound))
        self.assertTrue(reporter.is_exception_type_ignored(Unauthorized))
        self.assertTrue(reporter.is_exception_type_ignored(Redirect))

        os.environ['RAVEN_ENABLE_EXCEPTIONS'] = 'Unauthorized, Redirect'
        self.assertTrue(reporter.is_exception_type_ignored(NotFound))
        self.assertFalse(reporter.is_exception_type_ignored(Unauthorized))
        self.assertFalse(reporter.is_exception_type_ignored(Redirect))

    def test_no_exception_when_buildout_root_invalid(self):
        os.environ['RAVEN_BUILDOUT_ROOT'] = '/invalid/buildout/root'
        self.assertIsNone(reporter.get_release())



from zope.i18nmessageid import MessageFactory
from ftw.raven.monkeypatches import install_patches


_ = MessageFactory('ftw.raven')


def initialize(context):
    """Initializer called when used as a Zope 2 product.

    This is referenced from configure.zcml. Regstrations as a "Zope 2 product"
    is necessary for GenericSetup profiles to work, for example.

    """
    install_patches()



from AccessControl.users import nobody
from Acquisition import aq_acquire
from ftw.raven.client import get_raven_client
from ftw.raven.config import get_raven_config
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution
from plone.memoize import forever
from raven import fetch_git_sha
from raven.exceptions import InvalidGitRepository
from yolk.yolklib import Distributions
import logging
import sys


LOG = logging.getLogger('ftw.raven.reporter')


def maybe_report_exception(context, request, exc_type, exc, traceback):
    if is_exception_type_ignored(exc_type):
        return

    exc_info = exc_type, exc, traceback
    try:
        client = get_raven_client()
        if client is None:
            return

        try:
            data = {'request': prepare_request_infos(request),
                    'user': prepare_user_infos(context, request),
                    'extra': prepare_extra_infos(context, request),
                    'modules': prepare_modules_infos(),
                    'tags': prepare_tags(exc)}
            release = get_release()
            if release:
                data['release'] = release
        except:
            LOG.error('Error while preparing sentry data.')
            raise

        try:
            client.captureException(exc_info=exc_info, data=data)
        except:
            LOG.error('Error while reporting to sentry.')
            raise

    except:
        if context:
            aq_acquire(context, 'error_log').raising(sys.exc_info())
        try:
            get_raven_client().captureException(
                data={'extra': {
                    'raven_meta_error': 'Error occured while reporting'
                    ' another error.'}})
        except:
            LOG.error('Failed to report error occured while reporting error.')


def is_exception_type_ignored(exc_type):
    if exc_type is None:
        return True
    ignored_classnames = get_raven_config().ignored_exception_classnames
    return exc_type.__name__ in ignored_classnames


def prepare_request_infos(request):
    headers = {}
    environ = {}
    for key, value in request.environ.items():
        if key.upper().startswith('HTTP'):
            # HTTP_USER_AGENT -> User-Agent
            key = key[5:].replace('_', '-').title()
            headers[key] = value
        else:
            environ[key] = value

    return {
        'url': request.URL,
        'method': request.method,
        'data': request.form.copy(),
        'query_string': request.QUERY_STRING,
        'cookies': request.cookies.copy(),
        'headers': headers,
        'env': environ,
    }


def prepare_user_infos(context, request, include_roles=True):
    user = request.get('AUTHENTICATED_USER', nobody)

    data = {'ip_address': get_client_ip_address(request)}
    if include_roles:
        data['roles'] = user.getRoles()
        data['roles_in_context'] = user.getRolesInContext(context)

    if user and user != nobody:
        data.update({'id': user.getId(),
                     'username': user.getUserName(),
                     'email': user.getProperty('email', ''),
                     'fullname': user.getProperty('fullname', ''),
                     'authentication': 'authenticated'})
    else:
        data.update({'authentication': 'anonymous'})

    return data


def get_client_ip_address(request):
    """Returns the IP address of the host which sent the
    request initally.
    """
    ips = request.environ.get(
        'HTTP_X_FORWARDED_FOR',
        request.environ.get('REMOTE_ADDR'))

    if ips is None:
        return ''
    elif not hasattr(ips, 'split'):
        return ips
    else:
        return ips.split(',')[0].strip()


def prepare_modules_infos():
    dists = (dist for (dist, active)
             in Distributions().get_distributions('all')
             if active)
    modules = dict((dist.project_name, dist.version) for dist in dists)
    modules['python'] = '.'.join(map(str, sys.version_info[:3]))
    return modules


def prepare_tags(exc):
    tags = get_default_tags()
    if hasattr(exc, 'error_log_id'):
        tags['error_log_id'] = exc.error_log_id
    return tags


@forever.memoize
def get_default_tags():
    return get_raven_config().tags.copy()


def get_release():
    config = get_raven_config()
    if config.project_dist:
        try:
            return get_distribution(config.project_dist).version
        except DistributionNotFound:
            LOG.error('The distribution "{}" could not be found.'.format(
                config.project_dist))
            return None

    if config.buildout_root:
        try:
            return fetch_git_sha(config.buildout_root)
        except InvalidGitRepository:
            LOG.error('The path {} does not exist or is not'
                      ' a git repository'.format(config.buildout_root))
            return None

    return None


def prepare_extra_infos(context, request):
    return {'context': context,
            'request.other': request.other.copy()}



from App.ZApplication import ZApplicationWrapper
from ftw.raven.reporter import maybe_report_exception
from Products.SiteErrorLog.SiteErrorLog import SiteErrorLog
from ZODB.POSException import ConflictError
from ZPublisher.Publish import Retry
from Zope2.App.startup import ZPublisherExceptionHook
import re
import sys


def exception_hook_wrapper(self, published, REQUEST, t, v, traceback):
    report_exception = True
    args = [published, REQUEST, t, v, traceback]

    try:
        try:
            return self.__ori_call__(*args)

        except Retry:
            # There was probably a conflict, resulting in a Retry
            # and the request will be done once again completely.
            report_exception = False
            raise

        except ConflictError:
            # This request was probably executed 3 times
            # (with Retry-exceptions, see above) but could not be
            # finished because of a ConflictError.
            # We want to report the ConflictError now, that's why
            # we replace the previous exception (which was probably
            # a Retry exception) with our ConflictError.
            args[-3:] = sys.exc_info()
            report_exception = True
            raise

    finally:
        if report_exception:
            maybe_report_exception(*args)


def ZApplicationWrapper__repr__(self):
    """ZApplicationWrapper has no __repr__ because it does not inherit
    from object.
    For having stack locals in sentry, the raven client tries to make
    repr() for each object, which fails with the ZApplicationWrapper
    because there is no implementation.
    Therefore we add the __repr__ to the ZApplicationWrapper class.
    """
    mod = self.__class__.__module__
    cls = self.__class__.__name__
    mem = '0x' + hex(id(self))[2:].zfill(8).upper()
    return '<{0}.{1} instance at {2}>'.format(mod, cls, mem)


def SiteErrorLog_raising_wrapper(self, info):
    """The default error_log assigns random, not guessable IDs to errors.
    The return value of ``raising`` is usually the URL to the error_log
    entry identified by its ID.
    In order to be able to report the error_log ID to sentry later,
    we store it on the exception.
    """
    result = self.raising_original(info)
    if not info[1] or not isinstance(result, (str, unicode)):
        return result

    match = re.match('.*error_log/showEntry\?id=([\d\.]*)$', result)
    if match:
        info[1].error_log_id = match.group(1)

    return result


def install_patches():
    ZPublisherExceptionHook.__ori_call__ = ZPublisherExceptionHook.__call__
    ZPublisherExceptionHook.__call__ = exception_hook_wrapper
    ZApplicationWrapper.__repr__ = ZApplicationWrapper__repr__
    if SiteErrorLog.raising != SiteErrorLog_raising_wrapper:
        SiteErrorLog.raising_original = SiteErrorLog.raising
        SiteErrorLog.raising = SiteErrorLog_raising_wrapper



from ftw.raven.client import get_raven_client
from ftw.raven.client import purge_raven_client
from ftw.raven.testing import RAVEN_FUNCTIONAL
from ftw.raven.tests.client_mock import ClientMock
from ftw.testbrowser import Browser
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from unittest2 import TestCase
import transaction


class FunctionalTestCase(TestCase):
    layer = RAVEN_FUNCTIONAL

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        self.load_zcml_string = self.layer['load_zcml_string']
        ClientMock.install()

    def tearDown(self):
        purge_raven_client()

    def grant(self, *roles):
        setRoles(self.portal, TEST_USER_ID, roles)
        transaction.commit()

    def request_to_error_view(self, view='make_key_error', data=None,
                              login_as=None):
        # We need to make sure that the raven client mock is created
        # in this thread and not in the sub-thread created by the
        # testbrowser, in order to be able to read it later.
        get_raven_client()

        with Browser()(self.layer['app']) as browser:
            if login_as:
                browser.login(login_as)

            # We trigger errors on purpose here, so the testbrowser
            # should not raise on failed http requests.
            browser.raise_http_errors = False
            browser.open(view=view, data=data)



from ftw.upgrade import UpgradeStep


class SeperateLibAndConfigJavascripts(UpgradeStep):
    """Seperate lib and config javascripts.
    """

    def __call__(self):
        self.install_upgrade_profile()
